DATABASES = {'default': {
  'ENGINE': 'django.db.backends.sqlite3',
  'NAME': '.testi.sqlite',
}}
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
INSTALLED_APPS = ['lume', 'testit']
MIDDLEWARE = []
SECRET_KEY = 'epäjärjestelmällistyttämättömyydellänsäkäänköhän'
USE_TZ = True
